﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using AsReaderP52N;
using System.IO;
using System.Threading;

namespace AsReaderP52N_Demo
{
    public partial class Upgrade : Form
    {
        UpgradeType CurUpgradeType;
        byte[] fileBytes = { };
        bool CanUpgrade = false;
        UInt32 PackageNo = 0;
        bool isUpdating = false;
        public delegate void CloseUpdateWindow(bool isRestart);
        public CloseUpdateWindow closeUpdateWindow = null;
        public delegate void CloseUpdateWindowWithNothing();
        public CloseUpdateWindowWithNothing closeUpdateWindowWithNothing = null;
        bool isDisconnected = false;
        bool doUpdated = false;
        public ConnectType connectType { set; get; }
        public byte rs485Address { get; set; }
        //线程通知
        private static AutoResetEvent resetEvent = new AutoResetEvent(false);
        public Upgrade(UpgradeType upgradeType)
        {
            CurUpgradeType = upgradeType;
            InitializeComponent();
        }
        private void Form1_Load(object sender, EventArgs e)
        {
            lblProgressRate.Text = "0%";
            pbUpgrade.Value = 0;
#if DEBUG
            txtFilePath.Text = @"D:\guoxu\R2000_RF8008_4S4B_V1.2.3.bin";
#endif
            if (CurUpgradeType == UpgradeType.UpgradeApp) this.Text = "Upgrade App";
            else this.Text = "Upgrade Baseband";
            openFileDialog1 = new OpenFileDialog();
            //openFileDialog1.InitialDirectory = "c:\\";
            openFileDialog1.Filter = "bin files|*.bin|All files|*.*"; 
            openFileDialog1.RestoreDirectory = true;
            Device.whenReceivedData = ReceivedData;
            Device.whenDisconnect += ReceivedDisconnectInUpgrade;
            //Device.whenConnectError += new Device.WhenConnectError(ReceivedConnectErrorInUpgrade);
            //Device.SetUpgradeDelegates();
        }
        void ReceivedDisconnectInUpgrade()
        {
            isDisconnected = true;
            isUpdating = false;
            if (!this.IsHandleCreated)
            {
                Console.WriteLine("在创建窗口句柄之前，不能在控件上调用 Invoke 或 BeginInvoke。");
                return;
            }
            this.Invoke((Action)(() =>
            {
                MessageBox.Show("The remote connection has been disconnected.", "Tips", MessageBoxButtons.OK, MessageBoxIcon.Information);
            }));
            closeUpdateWindow?.Invoke(false);
            //}
        }
        private void btnSelect_Click(object sender, EventArgs e)
        {
            if (isUpdating) return;
            if (openFileDialog1.ShowDialog() == DialogResult.OK)
            {
                txtFilePath.Text = openFileDialog1.FileName;
            }
        }
        string msg = "You can restarting the device and upgrading again.";
        int failCnt = 0;
        private void btnStart_Click(object sender, EventArgs e)
        {
            if (isDisconnected)
            {
                MessageBox.Show("The device has been disconnected!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }
            if (isUpdating)
            {
                MessageBox.Show("It is upgrading!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }
            if (txtFilePath.Text == "")
            {
                MessageBox.Show("Select a bin File!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }
            if (!File.Exists(txtFilePath.Text))
            {
                MessageBox.Show("File not exists!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                return;
            }
            fileBytes = File.ReadAllBytes(txtFilePath.Text);
            if (fileBytes == null || fileBytes.Length == 0) return;
            isUpdating = true;
            doUpdated = true;
            lblProgressRate.Text = "0%";
            pbUpgrade.Value = 0;
            pbUpgrade.Maximum = (int)Math.Ceiling(fileBytes.Length / 256.0);
            UInt32 pckNo = 0;
            dicUpdatePkgResultCd.Clear();
            Task.Run(() =>
            {
                for (int i = 0; i < 3; i++)
                {
                    if (isDisconnected) return;
                    if (CurUpgradeType == UpgradeType.UpgradeBaseband) Device.UpgradeBaseband(pckNo, null, connectType, rs485Address);
                    else Device.UpgradeApp(pckNo, null, connectType, rs485Address);
                    resetEvent.WaitOne(3000);
                    if (dicUpdatePkgResultCd.ContainsKey(0) && dicUpdatePkgResultCd[0] == 0) break;
                    else if (i == 2 &&
                        (!dicUpdatePkgResultCd.ContainsKey(0) || dicUpdatePkgResultCd[0] != 0))
                    {
                        this.Invoke((Action)(() =>
                        {
                            failCnt++;
                            if (failCnt > 1) MessageBox.Show("Upgrade Error!" + msg, "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                            else MessageBox.Show("Upgrade Error!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                        }));
                        isUpdating = false;
                        return;
                    }
                    else continue;
                }
                if (CanUpgrade)
                {
                    failCnt = 0;
                    Thread thread = new Thread(DoUpdate);
                    thread.Start();
                }
                else
                {
                    this.Invoke((Action)(() =>
                    {
                        MessageBox.Show("Upgrade failed!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }));
                    isUpdating = false;
                }
            });
        }
        Dictionary<UInt32, int> dicUpdatePkgResultCd = new Dictionary<UInt32, int>();
        int failCntBin = 0;
        void DoUpdate()
        {
            try
            {
                PackageNo = 0;
                int sumFileSize = 0;// fileBytes.Length;
                int transferDataLength = 256;//一次传输的字节数
                int count = (int)Math.Ceiling(fileBytes.Length / 256.0);
                while (sumFileSize < fileBytes.Length)
                {
                    if (isDisconnected) return;
                    int remainder = fileBytes.Length - sumFileSize;
                    if (remainder < transferDataLength)
                    {
                        transferDataLength = remainder;
                    }
                    PackageNo++;
                    byte[] tmp = new byte[transferDataLength];
                    Array.Copy(fileBytes, sumFileSize, tmp, 0, transferDataLength);
                    sumFileSize += transferDataLength;
                    for (int i = 0; i < 20; i++)
                    {
                        if (isDisconnected) return;
                        if (CurUpgradeType == UpgradeType.UpgradeBaseband) Device.UpgradeBaseband(PackageNo, tmp, connectType, rs485Address);
                        else Device.UpgradeApp(PackageNo, tmp, connectType, rs485Address);
                        if (i == 0) resetEvent.WaitOne(3000);
                        else resetEvent.WaitOne(1000);
                        if (dicUpdatePkgResultCd.ContainsKey(PackageNo) && dicUpdatePkgResultCd[PackageNo] == 0) break;
                        else if (i == 19 &&
                            (!dicUpdatePkgResultCd.ContainsKey(PackageNo) || dicUpdatePkgResultCd[PackageNo] != 0))
                        {
                            this.Invoke((Action)(() =>
                            {
                                failCntBin++;
                                if (failCntBin > 1) MessageBox.Show("Communication timeout!" + msg, "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                                else MessageBox.Show("Communication timeout!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                            }));
                            isUpdating = false;
                            return;
                        }
                        else continue;
                    }
                    if (PackageNo > 0 && PackageNo == count)
                    {
                        PackageNo = Convert.ToUInt32("FFFFFFFF", 16);
                        for (int i = 0; i < 5; i++)
                        {
                            if (isDisconnected) return;
                            if (CurUpgradeType == UpgradeType.UpgradeBaseband) Device.UpgradeBaseband(PackageNo, null, connectType, rs485Address);
                            else Device.UpgradeApp(PackageNo, null, connectType, rs485Address);
                            resetEvent.WaitOne(3000);
                            if (dicUpdatePkgResultCd.ContainsKey(PackageNo) && dicUpdatePkgResultCd[PackageNo] == 0) break;
                            else if (i == 4 &&
                                (!dicUpdatePkgResultCd.ContainsKey(PackageNo) || dicUpdatePkgResultCd[PackageNo] != 0))
                            {
                                this.Invoke((Action)(() =>
                                {
                                    MessageBox.Show("File check error!", "Warning", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                                }));
                                isUpdating = false;
                                return;
                            }
                            else continue;
                        }
                    }
                }
                dicUpdatePkgResultCd.Clear();
                isUpdating = false;
            }
            catch (Exception ex)
            {
                Console.WriteLine("DoUpdate: " + ex.Message);
            }
        }
        void ReceivedData(byte[] data)
        {
            if (!isUpdating) return;
            try
            {
                byte resultCode;
                UInt32 pkgNo;
                if (data.Length > 9 && data[0] == 0x5A && data[5] == 0x00 && data[6] == 0x05)//数据长度区：0005
                {
                    if (data[1] == 0x00 && data[2] == 0x01 &&
                        ((connectType == ConnectType.RS485 && data[3] == 0x24) || (connectType != ConnectType.RS485 && data[3] == 0x04)) &&
                        (CurUpgradeType == UpgradeType.UpgradeBaseband && data[4] == 0x01) || (CurUpgradeType == UpgradeType.UpgradeApp && data[4] == 0x00))
                    {
                        Console.WriteLine("Upgrade ReceivedData: " + BitConverter.ToString(data));
                        byte[] tmp = data.Skip(1).Take(data.Length - 3).ToArray();
                        ushort checkCode = Device.CRC16_CCITT(tmp, tmp.Length);
                        ushort check = (ushort)((data[data.Length - 2] << 8 & 0xFF00) + data[data.Length - 1]);
                        if (checkCode != check)
                        {
                            Console.WriteLine("Upgrade ReceivedData: Frame check error.");
                            return;
                        }
                        if (!BitConverter.IsLittleEndian) pkgNo = BitConverter.ToUInt32(data, 7);
                        else pkgNo = BitConverter.ToUInt32(new byte[] { data[10], data[9], data[8], data[7] }, 0);
                        resultCode = data[11];
                        Console.WriteLine(DateTime.Now.ToString("HH:mm:ss.fff") + "Received UpgradeResult:" + resultCode.ToString() + "-" + pkgNo.ToString());
                        if (resultCode == 0)
                        {
                            if (dicUpdatePkgResultCd.ContainsKey(pkgNo))
                            {
                                if (dicUpdatePkgResultCd[pkgNo] != 0) dicUpdatePkgResultCd[pkgNo] = 0;
                            }
                            else
                            {
                                dicUpdatePkgResultCd.Add(pkgNo, 0);
                            }
                            if (pkgNo == 0)
                            {
                                CanUpgrade = true;
                                resetEvent.Set();
                            }
                            else
                            {
                                if (pkgNo == Convert.ToUInt32("FFFFFFFF", 16))
                                {
                                    resetEvent.Set();
                                    this.Invoke((Action)(() =>
                                    {
                                        failCnt = 0;
                                        failCntBin = 0;
                                        Device.whenDisconnect -= ReceivedDisconnectInUpgrade;
                                        if (CurUpgradeType == UpgradeType.UpgradeBaseband)
                                        {
                                            MessageBox.Show("Upgrade succeeded. The device will be restarted.", "Tips", MessageBoxButtons.OK);
                                            Device.Restart();
                                            closeUpdateWindow?.Invoke(true);
                                        }
                                        else
                                        {
                                            MessageBox.Show("Upgrade succeeded. You can restart the device manually.", "Tips", MessageBoxButtons.OK);
                                            closeUpdateWindow?.Invoke(false);
                                        }
                                        //else Device.Restart();
                                    }));
                                }
                                else
                                {
                                    resetEvent.Set();
                                    StartProgressBar(pkgNo);
                                }
                            }
                        }
                        else
                        {
                            if (!dicUpdatePkgResultCd.ContainsKey(pkgNo))
                            {
                                dicUpdatePkgResultCd.Add(pkgNo, resultCode);
                            }
                            resetEvent.Set();
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("Upgrade ReceivedData: " + ex.Message);
            }
        }
        private void StartProgressBar(object obj)
        {
            this.Invoke(
                new Action(delegate
                {
                    UInt32 value = (UInt32)obj;
                    if (value > 0)
                    {
                        int tmp = (int)(value * 100 / pbUpgrade.Maximum);
                        pbUpgrade.Value = (int)value;
                        pbUpgrade.Refresh();
                        lblProgressRate.Text = tmp + "%";
                        lblProgressRate.Refresh();
                    }
                }));
        }

        private void Upgrade_FormClosing(object sender, FormClosingEventArgs e)
        {
            Device.whenDisconnect -= ReceivedDisconnectInUpgrade;
            if (isUpdating && isDisconnected == false)
            {
                //MessageBox.Show("Can not close it for upgrading!", "Warning");
                //e.Cancel = true;
                if (MessageBox.Show("It is upgrading,do you want to stop upgrade?", "Tips", MessageBoxButtons.YesNo) == DialogResult.Yes)
                {
                    e.Cancel = false;
                }
                else
                {
                    e.Cancel = true;
                }
                isUpdating = false;
            }
            if (!doUpdated && isDisconnected == false)
            {
                closeUpdateWindowWithNothing?.Invoke();
            }
        }
    }
}
